/*--------------------------------------------------
v1.0 MSoft 14.05.2018
ESP-xx - 1MBit-Flash: noSPIFFS, DIO
IDE 1.8.5 - ESP 2.4.2 
--------------------------------------------------*/

#include <ESP8266WiFi.h>
#include <Ticker.h>
#include <AsyncMqttClient.h>
#include <ArduinoOTA.h>

//#define DEBUG
//#define MQTTDEBUG

#define HOSTNAME "OTA-SONOFF-4"            

#define led_wlan    4
#define relais_off  5
#define relais_on  12
#define keyPin     14

#define WIFI_SSID "xxxxxxx"
#define WIFI_PASSWORD "xxxxxxxxx"

#define MQTT_HOST IPAddress(192, 168, 0, 99)
#define MQTT_PORT 1883

// WiFi
WiFiEventHandler wifiConnectHandler;
WiFiEventHandler wifiDisconnectHandler;
Ticker wifiReconnectTimer;

// MQTT
AsyncMqttClient mqttClient;
Ticker mqttReconnectTimer;

Ticker key_timer;
void keyCheck(void);
byte key_count = 0;              // Anzahl Tickerläufe für Entprellung
bool rel_status = false;         // Relais-Status

Ticker led_timer;
void ledBlink(void);
bool led_status = false; 

void connectToMqtt(void);
void onMqttConnect(bool sessionPresent);
void onMqttDisconnect(AsyncMqttClientDisconnectReason reason);
void onMqttSubscribe(uint16_t packetId, uint8_t qos);
void onMqttUnsubscribe(uint16_t packetId);
void onMqttMessage(char* topic, char* payload, AsyncMqttClientMessageProperties properties, size_t len, size_t index, size_t total);
void onMqttPublish(uint16_t packetId);

void setup()
{
  digitalWrite(relais_off,LOW);
  pinMode(relais_off,OUTPUT);
  digitalWrite(relais_on,LOW);
  pinMode(relais_on,OUTPUT);
  digitalWrite(led_wlan,LOW);
  pinMode(led_wlan,OUTPUT);
  pinMode(keyPin,INPUT_PULLUP);

#ifdef DEBUG
  Serial.begin(115200);
  Serial.println("\r\nReset");
#endif
  
// Start WiFi
  wifiConnectHandler = WiFi.onStationModeGotIP(onWifiConnect);
  wifiDisconnectHandler = WiFi.onStationModeDisconnected(onWifiDisconnect);
  connectToWifi();

// Start MQTT server.
  mqttClient.onConnect(onMqttConnect);
  mqttClient.onDisconnect(onMqttDisconnect);
  mqttClient.onMessage(onMqttMessage);
  mqttClient.setClientId("ESP-SONOFF-4");
  mqttClient.setServer(MQTT_HOST, MQTT_PORT);
  mqttClient.setWill("SONOFF-4/Status",0,false,"offline");
  
// Start OTA server.
  ArduinoOTA.setHostname(HOSTNAME);
  ArduinoOTA.begin();

  key_timer.attach_ms(20, keyCheck);
  led_timer.attach_ms(500, ledBlink);
}

void loop()
{
  ArduinoOTA.handle();                      // Handle OTA server.
}

// ----------------------------------------------------------------------
void ledBlink()
{
  if (led_status)
  {
    digitalWrite(led_wlan, 0);
  }
  else
  {     
    digitalWrite(led_wlan, 1);
  }
  led_status = !led_status;  
}  

// ----------------------------------------------------------------------

void keyCheck()
{
  if (digitalRead(keyPin) == 0)
  {
/*
#ifdef MQTTDEBUG
        mqttClient.publish("SONOFF-4/Test Taster",1,false,"test");
#endif
*/
    key_count++;                           // Entprellung
    if (key_count == 4)                    // 80ms
    {
      if (rel_status)                      // Relais ein?
      {      
        digitalWrite(relais_off,LOW);
        rel_status = false;       
        mqttClient.publish("SONOFF-4/Hand",1,false,"aus");
      }
      else
      {      
        digitalWrite(relais_off,HIGH);
        rel_status = true;
        mqttClient.publish("SONOFF-4/Hand",1,false,"ein");
      }     
      key_count = 5;                       // Taster noch gedrückt 
    }
  }
  else
  {
    key_count = 0;                         // Taste losgelassen  
  }   
}

// ###############################################################
// ---------------------- WiFi-Tools ------------------------------
// ###############################################################

void connectToWifi()
{
#ifdef DEBUG
  Serial.println("Connecting to Wi-Fi...");
#endif

  WiFi.mode(WIFI_STA);
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
}

// ------------------------------------------------------------------------

void onWifiConnect(const WiFiEventStationModeGotIP& event)
{
#ifdef DEBUG
  Serial.println("Connected to Wi-Fi.");
  Serial.println(WiFi.localIP());
#endif

  led_timer.detach();
  digitalWrite(led_wlan,HIGH);
  connectToMqtt();
}

void onWifiDisconnect(const WiFiEventStationModeDisconnected& event)
{
#ifdef DEBUG
  Serial.println("Disconnected from Wi-Fi.");
#endif

  led_timer.attach_ms(500, ledBlink);
  digitalWrite(led_wlan,LOW);
  mqttReconnectTimer.detach(); // ensure we don't reconnect to MQTT while reconnecting to Wi-Fi
  wifiReconnectTimer.once(2, connectToWifi);
}

// ###############################################################
// ---------------------- MQTT-Tools ------------------------------
// ###############################################################

void connectToMqtt()
{
#ifdef DEBUG
  Serial.println("Connecting to MQTT...");
#endif

  mqttClient.connect();
}

void onMqttConnect(bool sessionPresent)
{
  uint16_t packetIdSub1 = mqttClient.subscribe("SONOFF-4/#", 0);

  String output = String(WiFi.RSSI(), DEC);

#ifdef DEBUG
  Serial.print("RSSI: ");
  Serial.println(output);
#endif

  mqttClient.publish("SONOFF-4/RSSI",1,false,output.c_str());
  mqttClient.publish("SONOFF-4/Status",1,false,"online");
}

void onMqttDisconnect(AsyncMqttClientDisconnectReason reason)
{
#ifdef DEBUG
  Serial.println("Disconnected from MQTT.");
#endif

  if (WiFi.isConnected()) {
    mqttReconnectTimer.once(2, connectToMqtt);
  }
}

void onMqttMessage(char* topic, char* payload, AsyncMqttClientMessageProperties properties, size_t len, size_t index, size_t total)
{
  String instring;
  for (int i = 0; i < len; i++)
  {
    instring += (char)payload[i];                  // erstmal den payload in einen String
  }

  instring.trim();
  String topicstring = topic;

#ifdef DEBUG
  Serial.print("Topic: ");
  Serial.println(topic);
  Serial.print("Payload: ");
  Serial.println(instring); 
#endif

  String topicarray[5];
  int i = 0, start = 0, topic_index = 0;
  
  while ((topic_index = topicstring.indexOf('/',start)) != -1)            // Slash gefunden           
  {
    topicarray[i] = topicstring.substring(start,topic_index);
    start = topic_index + 1;
    i++;
  }
  topicarray[i] = topicstring.substring(start);
  topicarray[i].trim();
  
  if (topicarray[1] == "Dose")                     
  {
    if ((instring == "ein") and !rel_status)
    {
      digitalWrite(relais_off,HIGH);
      rel_status = true;
      mqttClient.publish("SONOFF-4/Relais",1,false,"ein");
    }
    else if ((instring == "aus") and rel_status)
    {
      digitalWrite(relais_off,LOW);
      rel_status = false;
      mqttClient.publish("SONOFF-4/Relais",1,false,"aus");
    }
  } 
}



